"""
Validate and decode AWS SNS/SQS message
"""
from django.conf import settings
import urllib3
import json
import re
from base64 import b64decode
from aws_message.crypto import Signature, CryptoException


class SNSException(Exception):
    pass


class SNS(object):
    """
    Amazon Web Serice SNS Message Class
    """

    _message = None

    def __init__(self, message):
        """
        Amazon Web Service SNS Message object

        Takes an object representing the contents of an AWS SNS message

        Raises SNSException
        """
        self._message = message

    def validate(self):
        t = self._message['SignatureVersion']
        if t != '1':
            raise SNSException('Unknown SNS Signature Version: ' + t)

        sig_conf = {
            'cert': {
                'type': 'url',
                'reference': self._message['SigningCertURL']
            }
        }

        try:
            Signature(sig_conf).validate(
                self._signText(), b64decode(self._message['Signature']))
        except CryptoException as err:
            raise SNSException(
                '%s validation fail: %s' % (self._message['Type'], err))
        except Exception as err:
            raise SNSException(
                'Invalid SNS %s: %s' % (self._message['Type'], err))

    def _signText(self):
        to_sign = ''
        # see: ("http://docs.amazonwebservices.com/sns/latest/"
        #       "gsg/SendMessageToHttp.verify.signature.html")
        to_sign = self._sigElement('Message') + \
            self._sigElement('MessageId')
        if self._message['Type'] == 'Notification':
            if 'Subject' in self._message:
                to_sign += self._sigElement('Subject')

            to_sign += self._sigElement('Timestamp')
        elif self._message['Type'] == 'SubscriptionConfirmation':
            to_sign += self._sigElement('SubscribeURL')
            to_sign += self._sigElement('Timestamp')
            to_sign += self._sigElement('Token')

        to_sign += self._sigElement('TopicArn')
        to_sign += self._sigElement('Type')

        return to_sign.encode('utf-8')

    def _sigElement(self, el):
        return '%s\n%s\n' % (el, self._message[el])

    def extract(self):
        message = self._message['Message']

        if re.match(r'^[a-zA-Z0-9]+[=]{0,2}$', message):
            message = b64decode(message)

        return json.loads(message)

    def subscribe(self):
        try:
            http = urllib3.PoolManager(
                cert_reqs='CERT_REQUIRED',
                ca_certs=settings.AWS_CA_BUNDLE
            )
            r = http.request('GET', self._message['SubscribeURL'])
            if r.status != 200:
                raise SNSException(
                    'Subscribe to %s failure: status: %s' % (
                        self._message['TopicArn'], r.status))
        except urllib3.exceptions.HTTPError as err:
            raise SNSException(
                'Subscribe to %s failure: %s' % (
                    self._message['TopicArn'], err))



"""
Extract - Base class to pull body out of AWS message
"""
from aws_message.crypto import aes128cbc, Signature, CryptoException
from base64 import b64decode
import json


class ExtractException(Exception):
    pass


class Extract(object):
    """
    Base class to extract AWS message contents
    """
    def __init__(self, config, message):

        self._settings = config
        self._header = message['header']
        self._body = message['body']

    def extract(self):
        return self.parse(self._header['contentType'].lower(),
                          self._raw_message())

    def parse(self, content_type, body):
        if content_type == 'json':
            return json.loads(body)

        raise ExtractException('Unknown content-type: %s' % (content_type))

    def _raw_message(self):
        if self._settings.get('VALIDATE_MSG_SIGNATURE', True):
            self._validate()

        body = b64decode(self._body)

        try:
            if set(['keyId', 'iv']).issubset(self._header):
                key = self._header['keyId']
                keys = self._settings.get('KEYS', {})
                if key not in keys:
                    raise ExtractException('Invalid keyId : %s' (key))

                cipher = aes128cbc(b64decode(keys[key]),
                                   b64decode(self._header['iv']))
                body = cipher.decrypt(body)
        except CryptoException as err:
            raise ExtractException('Cannot decrypt: %s' % (err))
        except Exception as err:
            raise ExtractException('Cannot read: %s' % (err))

        return body

    def _validate(self):
        to_sign = self._header[u'contentType'] + '\n'
        if 'keyId' in self._header:
            to_sign += self._header[u'iv'] + '\n' + \
                       self._header[u'keyId'] + '\n'
        to_sign += self._header[u'messageContext'] + '\n' + \
            self._header[u'messageId'] + '\n' + \
            self._header[u'messageType'] + '\n' + \
            self._header[u'sender'] + '\n' + \
            self._header[u'signingCertUrl'] + '\n' + \
            self._header[u'timestamp'] + '\n' + \
            self._header[u'version'] + '\n' + \
            self._body + '\n'

        sig_conf = {
            'cert': {
                'type': 'url',
                'reference': self._header[u'signingCertUrl']
            }
        }

        try:
            Signature(sig_conf).validate(to_sign.encode('ascii'),
                                         b64decode(self._header['signature']))
        except CryptoException as err:
            raise ExtractException('Cannot unencode message: %s' % (err))
        except Exception as err:
            raise ExtractException('Invalid signature %s: %s' (
                self._header['signature'], err))



from django.conf import settings
from aws_message.aws import SNS, SNSException
from aws_message.sqs import SQSQueue
from logging import getLogger
import json


class GatherException(Exception):
    pass


class Gather(object):
    """
    Class to gather event messages from AWS SQS queue,
    validate and process their content
    """
    def __init__(self, sqs_settings=None, processor=None, exception=None):
        if not processor:
            raise GatherException('missing event processor')

        self._processor = processor
        self._exception = exception if exception \
            else processor.EXCEPTION_CLASS
        self._settings = sqs_settings if sqs_settings \
            else settings.AWS_SQS.get(processor.SETTINGS_NAME)
        self._topicArn = self._settings.get('TOPIC_ARN')
        self._queue = SQSQueue(settings=self._settings)
        self._log = getLogger(__name__)

    def gather_events(self):
        to_fetch = self._settings.get('MESSAGE_GATHER_SIZE')
        while to_fetch > 0:
            n = min([to_fetch, 10])
            msgs = self._queue.get_messages(
                num_messages=n,
                visibility_timeout=self._settings.get('VISIBILITY_TIMEOUT'))

            for msg in msgs:
                try:
                    sqs_msg = json.loads(msg.get_body())
                    if sqs_msg['TopicArn'] == self._topicArn:
                        raw_message = SNS(sqs_msg)

                        if self._settings.get('VALIDATE_SNS_SIGNATURE', True):
                            raw_message.validate()

                        if sqs_msg['Type'] == 'Notification':
                            settings = self._settings.get(
                                'PAYLOAD_SETTINGS', {})
                            message = raw_message.extract()
                            self._processor(settings, message).process()
                        elif sqs_msg['Type'] == 'SubscriptionConfirmation':
                            self._log.debug(
                                'SubscribeURL: ' + sqs_msg['SubscribeURL'])
                    else:
                        self._log.warning(
                            'Unrecognized TopicARN : ' + sqs_msg['TopicArn'])
                except ValueError as err:
                    raise GatherException('JSON : %s' % err)
                except self._exception as err:
                    raise GatherException("MESSAGE: %s" % err)
                except SNSException as err:
                    raise GatherException("SNS: %s" % err)
                except Exception as err:
                    self._log.exception("Gather Error")
                    raise GatherException("ERROR: %s" % err)
                else:
                    self._queue.delete_message(msg)

            if len(msgs) < n:
                self._log.debug("SQS drained")
                break

            to_fetch -= n
